#include <stdio.h>
#include <string.h>

// Dies ist ein einfacher MIMA-Bytecode Interpreter
// Thorsten Rapp, 01.06.2004
//
// kompilieren: gcc runmima.c
//      Aufruf: ./a.out programm.mic
//
// Das Format der Datei .mic sieht wie folgt aus:
// 0-4. Byte: MIMA\0     Erkenner fr MIMA Bytecode
// 5-7. Byte: 0x??????   IAR Anfangsinhalt
// 8. Byte: Anzahl von Eingabefelder
// 9. Byte: Anzahl von Ausgabefelder
// Liste von Eingabezellenadressen
// Liste von Ausgabezellenadressen
// Speicherinhalt ab Zelle 0 bis nach Bedarf.

int mem[0x0FFFFF];  //A0-A19=2^20 Adressen  3 Byte
int IAR;
int Akku;
int IR;
int print[256];
int ask[256];

main(int argc, char *argv[])
{
  printf("MIMA - Bytecode Interpreter (c) Technische Informatik II\n\n");

  if (!(argv[1]))
  {
	printf("Aufruf: mima programm.mic\n");
	exit(1);
  }

// einlesen:
// Der Dateieinleseteil ist zweckmig aber nicht schn programmiert.
// Sorry, aber die Hauptkonzentration war den MIMA Programmen selbst
// gewidmet. Dieser Interpreter ist nur ein Bonus.

  FILE *ptr;

  int c1;
  int c2;
  int c3;

  ptr = fopen(argv[1],"rb");
  int i=0;

  char mimacheck[10];
  fgets(mimacheck,5,ptr);
  
  if ((!(strcmp(mimacheck,"MIMA"))==0)|(fgetc(ptr)!=0))
  {
	  printf("Kein MIMA Bytecode!\n");
	  exit(1);
  }

  IAR=fgetc(ptr)*256*256+fgetc(ptr)*256+fgetc(ptr);

  int inputlaenge=fgetc(ptr);
  int outputlaenge=fgetc(ptr);
 
  for (i=0;i<inputlaenge;i++)
  {
	  ask[i]=fgetc(ptr)*256*256+fgetc(ptr)*256+fgetc(ptr);
  }  
  
  ask[i]=-1;

  for (i=0;i<outputlaenge;i++)
  {
	  print[i]=fgetc(ptr)*256*256+fgetc(ptr)*256+fgetc(ptr);
  }  

  print[i]=-1;
  
  // mem[] herstellen
  i=0;
  while ((c1 = fgetc(ptr)) != EOF)
  {
   c2=fgetc(ptr);
   c3=fgetc(ptr);
   mem[i]=c1*256*256+c2*256+c3;
   i++;
  }
  
  fclose(ptr);



// ausfhren:

  for (i=0;ask[i]!=-1;i++)
  {
	  printf("Bitte Inhalt fr Zelle [0x%05X] eingeben: ",ask[i]);
	  scanf("%i",&mem[ask[i]]);
  }

int a;
while (a==a)
{

//Fetch-Phase:
IR=mem[IAR];
IAR++;

//Dekodierpase 
int OP1=IR&0xF00000;
OP1>>=20;
int OP2=IR&0x0F0000;
OP2>>=16;

switch (OP1)
{
  case(0): LDC();break;
  case(1): LDV();break;
  case(2): STV();break;
  case(3): ADD();break;
  case(4): AND();break;
  case(5): OR();break;
  case(6): XOR();break;
  case(7): EQL();break;
  case(8): JMP();break;
  case(9): JMN();break;
 case(10): LDIV();break;
 case(11): STIV();break;
 case(12): JMS();break;
 case(13): JIND();break;
 case(15): {
	         switch (OP2) 
		     { 
	           case(0): HALT();break;
			   case(1): NOT();break;
			   case(2): RAR();break;
			   default: FEHLER();break;
             }//switch OP2
		   break;
           }//case15

   default: FEHLER();break;
}//switch OP1
}//while


}//main.


//Ausfhrungsphase:

LDC ()
{
  Akku=IR&0xFFFFF;
}

LDV ()
{
  Akku=mem[IR&0xFFFFF];
}

STV ()
{
  mem[IR&0xFFFFF]=Akku;
}

ADD ()
{
  Akku+=mem[IR&0xFFFFF];
  Akku%=0x1000000;
}

AND ()
{
  Akku&=mem[IR&0xFFFFF];
}

OR ()
{
  Akku|=mem[IR&0xFFFFF];
}

XOR ()
{
  Akku^=mem[IR&0xFFFFF];
}

EQL ()
{
	  if (mem[IR&0xFFFFF]==Akku)
	  {
		  Akku=-1;
	  }
	  else
	  {
		  Akku=0;
	  }
}

JMP ()
{
  IAR=IR&0xFFFFF;
}

JMN ()
{
   if ((Akku&0x800000)!=0)
   {
	   IAR=IR&0xFFFFF;
   }
}

LDIV ()
{
   Akku=mem[mem[IR&0xFFFFF]];
}

STIV ()
{
   mem[mem[IR&0xFFFFF]]=Akku;
}

JMS ()
{
  mem[IR&0xFFFFF]=IAR;
  IAR=(IR&0xFFFFF)+1;
}

JIND ()
{
  IAR=mem[IR&0xFFFFF];
}

HALT ()
{
  printf("\nAusgabe der ausgesuchten Speicherzellen:\n");

  int i;
  for (i=0;print[i]!=-1;i++)
  {
	  printf("[0x%05X]: 0x%06X ; Dezimal:(%i)\n",print[i],mem[print[i]],mem[print[i]]);
  }

  printf("\n");
  exit(1);
}

NOT ()
{
    Akku=~Akku;
	Akku&=0xFFFFFF;
}

RAR ()
{
   int rein=(Akku&1);
   Akku>>=1;
   rein<<=23;
   Akku|=rein;
   Akku%=0x1000000;
}

FEHLER()
{
  printf("Fehler im Bytecode: Unbekannter OpCode!\n");
  exit(1);
}
